#  -*- coding: utf-8 -*-
# *****************************************************************************
# ufit, a universal scattering fitting suite
#
# Copyright (c) 2013-2014, Georg Brandl and contributors.  All rights reserved.
# Licensed under a 2-clause BSD license, see LICENSE.
# *****************************************************************************

"""Panel for mappings."""

from numpy import array, mgrid
from matplotlib.cbook import flatten

from PyQt4.QtCore import SIGNAL
from PyQt4.QtGui import QFrame, QMessageBox

from ufit.data.dataset import Dataset
from ufit.data.mapping import bin_mapping
from ufit.models.peaks import Gauss2D
from ufit.utils import attrdict
from ufit.gui import logger
from ufit.gui.common import loadUi
from ufit.gui.session import session, SessionItem


def maybe_float(text, default):
    try:
        return float(text)
    except ValueError:
        return default


default_settings = {
    'title': 'mapping',
    'usemask': True,
    'contour': True,
    'dots': True,
    'interp': 100,
    'xaxis': '',
    'yaxis': '',
    'yscale': 1,
    'zmin': -1e300,
    'zmax': 1e300,
    'logz': False,
    'gauss2d': True,
}


class MappingItem(SessionItem):

    itemtype = 'mapping'

    def __init__(self, datas, settings):
        self.datas = datas
        self.settings = attrdict(settings or default_settings)
        SessionItem.__init__(self)

    def __reduce__(self):
        return (self.__class__, (self.datas, self.settings))

    def create_panel(self, mainwindow, canvas):
        return MappingPanel(mainwindow, canvas, self)

    def update_htmldesc(self):
        self.title = self.settings.title
        self.htmldesc = '<img src=":/map.png">&nbsp;&nbsp;<big><b>%d</b></big> ' \
                        '- %s<br>%s' % (
            self.index, self.settings.title,
            ', '.join(d.name for d in self.datas))
        session.emit(SIGNAL('itemUpdated'), self)


class MappingPanel(QFrame):

    def __init__(self, parent, canvas, item):
        QFrame.__init__(self, parent)
        loadUi(self, 'mapping.ui')
        self._limits = None
        self.item = item
        self.logger = logger.getChild('mapping')
        self.canvas = canvas
        self.mapdata = None
        self.set_datas(item.datas)
        self.set_settings(item.settings)

    @property
    def title(self):
        return self.settings.title

    def set_index(self, index):
        self.index = index
        self.gen_htmldesc()

    def set_datas(self, datas):
        axes = set(colname[4:] for colname in datas[0].meta
                   if colname.startswith('col_'))
        for data in datas[1:]:
            axes &= set(colname[4:] for colname in data.meta
                        if colname.startswith('col_'))
        axes = sorted(axes)
        self.xaxisBox.addItems(axes)
        self.yaxisBox.addItems(axes)

    def set_settings(self, s):
        """Update controls from settings."""
        self.titleBox.setText(s.title)
        self.xaxisBox.setCurrentIndex(self.xaxisBox.findText(s.xaxis))
        self.yaxisBox.setCurrentIndex(self.yaxisBox.findText(s.yaxis))
        self.stepBox.setValue(s.interp)
        self.zminEdit.setText(str(s.zmin))
        self.zmaxEdit.setText(str(s.zmax))
        self.scaleEdit.setText(str(s.yscale))
        self.usemaskBox.setChecked(s.usemask)
        self.dotsBox.setChecked(s.dots)
        self.contourBox.setChecked(s.contour)
        self.logBox.setChecked(s.logz)
        self.fitBox.setChecked(s.gauss2d)

    def _update_settings(self):
        """Update settings from controls."""
        s = self.item.settings
        old_title = s.title
        title = s.title = self.titleBox.text()
        if title != old_title:
            self.item.update_htmldesc()
        xaxis = s.xaxis = str(self.xaxisBox.currentText())
        yaxis = s.yaxis = str(self.yaxisBox.currentText())
        if xaxis == yaxis:
            QMessageBox.warning(self, 'Error', 'Please select distinct X '
                               'and Y axes.')
            return
        s.interp = self.stepBox.value()
        s.zmin = maybe_float(self.zminEdit.text(), -1e300)
        s.zmax = maybe_float(self.zmaxEdit.text(), 1e300)
        s.yscale = maybe_float(self.scaleEdit.text(), 1.0)
        s.usemask = self.usemaskBox.isChecked()
        s.dots = self.dotsBox.isChecked()
        s.contour = self.contourBox.isChecked()
        s.logz = self.logBox.isChecked()
        s.gauss2d = self.fitBox.isChecked()
        session.set_dirty()

    def on_buttonBox_clicked(self, button):
        """Apply button clicked."""
        self._update_settings()
        self.rebuild_map(quiet=False)
        self.plot()

    def rebuild_map(self, quiet=True):
        s = self.item.settings
        try:
            self.mapdata = bin_mapping(s.xaxis, s.yaxis, self.item.datas,
                                       usemask=s.usemask, log=s.logz,
                                       yscale=s.yscale, interpolate=s.interp,
                                       minmax=(s.zmin, s.zmax))
        except Exception, err:
            self.logger.exception('While creating mapping')
            if not quiet:
                QMessageBox.warning(self, 'Mapping error',
                                    'Could not create mapping: %s (have you '
                                    'selected the right columns?)' % err)
            return

    def plot(self, limits=True, canvas=None):
        s = self.item.settings
        canvas = canvas or self.canvas
        canvas.plotter.reset(limits)
        if not s.xaxis:
            canvas.draw()
            return
        if not self.mapdata:
            self.rebuild_map()
        canvas.plotter.plot_mapping(
            s.xaxis, s.yaxis, self.mapdata, title=s.title,
            mode=int(s.contour), dots=s.dots)
        if s.gauss2d:
            self.fit_2dgauss(s.xaxis, s.yaxis)
        canvas.draw()

    def save_limits(self):
        self._limits = self.canvas.axes.get_xlim(), self.canvas.axes.get_ylim()

    def get_saved_limits(self):
        return self._limits

    def fit_2dgauss(self, x, y):
        runs = self.item.datas
        if self.item.settings.usemask:
            xss = array(list(flatten(run['col_'+x][run.mask] for run in runs)))
            yss = array(list(flatten(run['col_'+y][run.mask] for run in runs)))
            zss = array(list(flatten(run.y[run.mask] for run in runs)))
            dzss = array(list(flatten(run.dy[run.mask] for run in runs)))
        else:
            xss = array(list(flatten(run['col_'+x] for run in runs)))
            yss = array(list(flatten(run['col_'+y] for run in runs)))
            zss = array(list(flatten(run.y for run in runs)))
            dzss = array(list(flatten(run.dy for run in runs)))
        maxidx = zss.argmax()
        xdata = array((xss, yss)).T
        model = Gauss2D(pos_x=xss[maxidx], pos_y=yss[maxidx],
                        fwhm_x=0.5*(xss.max()-xss.min()),
                        fwhm_y=0.5*(yss.max()-yss.min()), ampl=zss[maxidx])
        data = Dataset.from_arrays('2dgauss', xdata, zss, dzss)
        res = model.fit(data)
        xx, yy = mgrid[xss.min():xss.max():100j,
                       yss.min():yss.max():100j]
        mesh = array((xx.ravel(), yy.ravel())).T
        zmesh = model.fcn(res.paramvalues, mesh).reshape((100, 100))
        ax = self.canvas.figure.gca()
        ax.contour(xx, yy, zmesh)
        self.fitParLbl.setText('pos_x: %(pos_x).5f  pos_y: %(pos_y).5f  '
                               'theta: %(theta).5f  '
                               'fwhm_x: %(fwhm_x).5f  fwhm_y: %(fwhm_y).5f  '
                               'ampl: %(ampl).5f' % res.paramvalues)



import os
from distutils.core import setup

def find_packages():
    """Return a list of all ufit subpackages."""
    out = ['ufit']
    stack = [('ufit', 'ufit.')]
    while stack:
        where, prefix = stack.pop(0)
        for name in os.listdir(where):
            fn = os.path.join(where, name)
            if '.' not in name and os.path.isdir(fn) and \
                    os.path.isfile(os.path.join(fn, '__init__.py')):
                out.append(prefix + name)
                stack.append((fn, prefix + name + '.'))
    return out

def find_ui_files():
    """Find all Qt .ui files in ufit.gui subpackages."""
    res = {}
    for root, dirs, files in os.walk('ufit/gui/ui'):
        uis = [uifile for uifile in files if uifile.endswith('.ui')]
        if uis:
            res[root.replace('/', '.')] = uis
    return res

import ufit

setup(
    name = 'ufit',
    version = ufit.__version__,
    license = 'GPL',
    author = 'Georg Brandl',
    author_email = 'georg.brandl@frm2.tum.de',
    description = 'Universal scattering data fitting tool',
    url = 'https://bitbucket.org/birkenfeld/ufit/',
    packages = find_packages(),
    package_data = find_ui_files(),
    scripts = ['ufitgui'],
)



#  -*- coding: utf-8 -*-
# *****************************************************************************
# ufit, a universal scattering fitting suite
#
# Copyright (c) 2013-2014, Georg Brandl and contributors.  All rights reserved.
# Licensed under a 2-clause BSD license, see LICENSE.
# *****************************************************************************

"""Loader for cascade image data."""

from numpy import frombuffer, sqrt


def check_data(fp):
    fp.read(128*128*4)  # cannot decide anything from the pixel data
    footerid = fp.read(18)
    fp.seek(0, 0)
    return footerid == '\n### NICOS Cascade'


def guess_norm(meta):
    maxmon = 0
    mg = None
    for k in meta:
        if k.startswith('mon') and k[3:].isdigit():
            if meta[k] > maxmon:
                maxmon = meta[k]
                mg = k
    return mg


def read_data(filename, fp):
    meta = {}

    arr = frombuffer(fp.read(128*128*4), '<I4').reshape((128, 128)).astype(float)
    darr = sqrt(arr)

    # now read metadata
    remark = ''
    for line in fp:
        if line.startswith('#'):
            continue
        items = line.strip().split(None, 2)
        try:
            oval, unit = items[2].split(None, 1)
            val = float(oval)
        except (IndexError, ValueError):
            try:
                oval = items[2]
                val = float(oval)
            except ValueError:
                val = items[2]
            except IndexError:
                continue
        key = items[0]
        if key.endswith(('_offset', '_precision')):
            # we don't need these for fitting
            continue
        if key.endswith('_value'):
            key = key[:-6]
        if key.endswith('_instrument'):
            meta['instrument'] = oval.lower()
            continue
        elif key.endswith('_proposal'):
            meta['experiment'] = oval.lower()
        elif key.endswith('_samplename'):
            meta['title'] = oval
        elif key.endswith('_remark'):
            remark = oval
        elif key == 'number':
            meta['filenumber'] = int(oval)
            continue
        elif key == 'info':
            meta['subtitle'] = val
            continue
        meta[key] = val
    if remark and 'title' in meta:
        meta['title'] += ', ' + remark

    return arr, darr, meta



#  -*- coding: utf-8 -*-
# *****************************************************************************
# ufit, a universal scattering fitting suite
#
# Copyright (c) 2013-2014, Georg Brandl and contributors.  All rights reserved.
# Licensed under a 2-clause BSD license, see LICENSE.
# *****************************************************************************

"""Fit result class."""

import copy

from numpy import array, linspace, ravel
from matplotlib import pyplot as pl

from ufit.utils import cached_property
from ufit.plotting import DataPlotter

__all__ = ['Result']


class Result(object):
    def __init__(self, success, data, model, params, message, chisqr):
        self.success = success
        self.data = data
        self.model = model
        self.params = params
        self.message = message
        self.chisqr = chisqr

    def __getitem__(self, key):
        return self.paramdict[key]

    def copy(self):
        return copy.deepcopy(self)

    @cached_property
    def paramdict(self):
        """Returns a dictionary mapping parameter names to parameter objects."""
        return dict((p.name, p) for p in self.params)

    @cached_property
    def paramvalues(self):
        """Returns a dictionary mapping parameter names to values."""
        return dict((p.name, p.value) for p in self.params)

    @cached_property
    def paramerrors(self):
        """Returns a dictionary mapping parameter names to errors."""
        return dict((p.name, p.error) for p in self.params)

    @cached_property
    def values(self):
        """Returns a list with all parameter values."""
        return [p.value for p in self.params]

    @cached_property
    def errors(self):
        """Returns a list with all parameter errors."""
        return [p.error for p in self.params]

    @cached_property
    def results(self):
        """Returns a list with the parameter values, then the parameter errors
        and the chi-square value concatenated.
        """
        return self.values + self.errors + [self.chisqr]

    @cached_property
    def residuals(self):
        """Returns the array of residuals."""
        return self.model.fcn(self.paramvalues, self.data.x) - self.data.y

    @cached_property
    def xx(self):
        """Returns a fine-spaced array of X values between the minimum and maximum
        of the original data X values.
        """
        return linspace(self.data.x.min(), self.data.x.max(), 1000)

    @cached_property
    def yy(self):
        """Returns the model evaluated at self.xx."""
        return self.model.fcn(self.paramvalues, self.xx)

    def printout(self):
        """Print out a table of the fit result and the parameter values.

        The chi-square value is also included in the table.  Example output::

           Fit results for 373
           ---------------------------------------------------------------------
           bkgd            =     5.5111 +/-    0.21427
           slope           =    -1.0318 +/-    0.16187
           inc_pos         =  -0.015615 +/- 0.00066617
           inc_ampl        =     547.21 +/-     8.0482
           inc_fwhm        =    0.12656 +/-  0.0012489
           dho_center      =  -0.015615 +/-          0 (fixed: inc_pos)
           dho_pos         =    0.41689 +/-  0.0086916
           dho_ampl        =    0.36347 +/-   0.034156
           dho_gamma       =    0.22186 +/-   0.025093
           dho_tt          =         16 +/-          0 (fixed: data.T)
           chi^2/NDF       =      1.491
           =====================================================================
        """
        print 'Fit results for %s' % self.data.name
        if not self.success:
            print 'FIT FAILED: ' + self.message
        elif self.message:
            print '>', self.message
        print '-' * 80
        for p in self.params:
            print p
        print '%-15s = %10.4g' % ('chi^2/NDF', self.chisqr)
        print '=' * 80

    def plot(self, axes=None, params=True, multi=False):
        """Plot the data and model together in the current figure.

        If *params* is true, also plot parameter values as text.
        """
        plotter = DataPlotter(axes=axes)
        c = plotter.plot_data(self.data, multi=multi)
        plotter.plot_model(self.model, self.data, paramvalues=self.paramvalues,
                           labels=not multi, color=c)
        if params and not multi:
            plotter.plot_params(self.params, self.chisqr)

    def plotfull(self, axes=None, params=True):
        """Plot the data and model, including subcomponents, together in the
        current figure or the given *axes*.

        If *params* is true, also plot parameter values as text.
        """
        plotter = DataPlotter(axes=axes)
        plotter.plot_data(self.data)
        plotter.plot_model_full(self.model, self.data,
                                paramvalues=self.paramvalues)
        if params:
            plotter.plot_params(self.params, self.chisqr)


def calc_panel_size(num):
    for nx, ny in ([1,1], [2,1], [2,2], [3,2], [3,3], [4,3], [5,3], [4,4],
                   [5,4], [6,4], [5,5], [6,5], [7,5], [6,6], [8,5], [7,6],
                   [9,5], [8,6], [7,7], [9,6], [8,7], [9,7], [8,8], [10,7],
                   [9,8], [11,7], [9,9], [11,8], [10,9], [12,8], [11,9], [10,10]):
        if nx*ny >= num:
            return nx, ny
    return num//10 + 1, 10


class MultiResult(list):
    def plot(self):
        dims = calc_panel_size(len(self))
        fig, axarray = pl.subplots(dims[1], dims[0])
        for res, axes in zip(self, ravel(axarray)):
            res.plotfull(axes=axes)
        #pl.tight_layout()

    @cached_property
    def datavalues(self):
        d = dict((k, [v]) for (k, v) in self[0].data.meta.iteritems())
        for res in self[1:]:
            for k, v in res.data.meta.iteritems():
                d[k].append(v)
        return d

    @cached_property
    def paramvalues(self):
        """Return a dictionary mapping parameter names to arrays of
        parameter values, one for each result.
        """
        d = dict((p.name, [p.value]) for p in self[0].params)
        for res in self[1:]:
            for p in res.params:
                d[p.name].append(p.value)
        for k in d:
            d[k] = array(d[k])
        return d

    @cached_property
    def paramerrors(self):
        """Return a dictionary mapping parameter names to arrays of
        parameter errors, one for each result.
        """
        d = dict((p.name, [p.error]) for p in self[0].params)
        for res in self[1:]:
            for p in res.params:
                d[p.name].append(p.error)
        for k in d:
            d[k] = array(d[k])
        return d

    def printout(self):
        """Print global parameters of the fit."""
        print 'OVERALL fit results'
        print '-' * 80
        for p in self[0].params:
            if p.overall:
                print p
        print '=' * 80

    def plot_param(self, xname, pname):
        pl.errorbar(self.datavalues[xname], self.paramvalues[pname],
                    self.paramerrors[pname], fmt='o-')
        pl.xlabel(xname)
        pl.ylabel(pname)



#  -*- coding: utf-8 -*-
# *****************************************************************************
# ufit, a universal scattering fitting suite
#
# Copyright (c) 2013-2014, Georg Brandl and contributors.  All rights reserved.
# Licensed under a 2-clause BSD license, see LICENSE.
# *****************************************************************************

"""
Utilities for clustering execution of a piece of code to multiple hosts using
SSH transport.
"""

import sys
import md5
import Queue
import threading
import cPickle as pickle
from os import path
from time import time

import paramiko

keyname = path.expanduser('~/.ufitcluster/key')
clusterlist = []

def init_cluster():
    try:
        fp = open(path.expanduser('~/.ufitcluster/hosts'))
        for line in fp: # user@host
            if line.startswith('#'):
                continue
            login, host = line.strip().split('@')
            clusterlist.append((login, host))
        fp.close()
    except Exception, err:
        print >>sys.stderr, 'Cannot read the list of cluster hosts:', err
        print >>sys.stderr, 'Please create a file %s with one line for each ' \
            'parallel process, for example:' % \
            path.expanduser('~/.ufitcluster/hosts')
        print >>sys.stderr
        print >>sys.stderr, 'username@hostname1'
        print >>sys.stderr, 'username@hostname1'
        print >>sys.stderr, 'username@hostname1'
        print >>sys.stderr, 'username@hostname1'
        print >>sys.stderr, 'username@hostname2'
        print >>sys.stderr, 'username@hostname2'
        print >>sys.stderr
        print >>sys.stderr, 'would run 4 parallel processes on hostname1 and ' \
            '2 parallel processes on hostname2.'
        print >>sys.stderr, 'For local operation only, use "localhost" as ' \
            'hostname.'
        print >>sys.stderr, 'The file %s must contain the SSH private key to ' \
            'use for the connection, even for localhost.' % \
            path.expanduser('~/.ufitcluster/key')


def client_runner(client, task_queue, result_queue):
    while True:
        code, funcname, jobnum, args = task_queue.get()
        if jobnum == -1:
            print '[C] exiting runner for %s' % client._host
            return
        try:
            sid = md5.new(str(time()) + str(args)).hexdigest()
            sys.stdout.write('.'); sys.stdout.flush()
            #print '[C] starting job on %s: %s' % (client._host, sid)
            code_footer = '''\nif __name__ == "__main__":
            import cPickle as pickle
            args = pickle.loads(%r)
            print pickle.dumps(%s(*args))
            ''' % (pickle.dumps(args), funcname)

            sftp = client.open_sftp()
            fobj = sftp.file('/tmp/ufit_cluster_%s.py' % sid, 'wb')
            fobj.write(code + code_footer)
            fobj.close()

            stdin, stdout, stderr = \
                client.exec_command('python /tmp/ufit_cluster_%s.py; '
                                    'rm /tmp/ufit_cluster_%s.py' % (sid, sid))
            result = pickle.load(stdout)
            #print '[C] done with job %s: %r' % (sid, result)
            result_queue.put((jobnum, result))
        except Exception, err:
            print '[C] no result on %s, requeuing: %r' % (client._host, err)
            task_queue.put((code, funcname, jobnum, args))
            return


clients = []
runners = []
task_queue = Queue.Queue()
result_queue = Queue.Queue()
cluster_setup = False

def setup_cluster():
    global cluster_setup
    if cluster_setup:
        return
    for i, (user, host) in enumerate(clusterlist):
        client = paramiko.SSHClient()
        client._host = '%s[%d]' % (host, i)
        client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        client.connect(host, username=user, key_filename=keyname,
                       look_for_keys=False)
        clients.append(client)
        print '[C] opened cluster connection to %s' % client._host
        runner = threading.Thread(target=client_runner,
            args=(client, task_queue, result_queue))
        runners.append(runner)
        runner.start()
    cluster_setup = True

def kill_cluster():
    for cl in clients:
        task_queue.put((None, None, -1, None))  # end!
        cl.close()


def run_cluster(code, funcname, argumentslist):
    setup_cluster()
    njobs = len(argumentslist)
    retval = [None] * njobs
    returns = 0
    for job in enumerate(argumentslist):
        task_queue.put((code, funcname) + job)
    while returns < njobs:
        # XXX check if any client is still running
        jobnum, result = result_queue.get()
        if isinstance(result, Exception):
            raise result
        retval[jobnum] = result
        returns += 1
    return retval

init_cluster()

if __name__ == '__main__':
    t1 = time()
    print run_cluster('def foo(a): return a\n', 'foo', [('a',),])
    t2 = time()
    print t2-t1



#  -*- coding: utf-8 -*-
# *****************************************************************************
# ufit, a universal scattering fitting suite
#
# Copyright (c) 2013-2014, Georg Brandl and contributors.  All rights reserved.
# Licensed under a 2-clause BSD license, see LICENSE.
# *****************************************************************************

"""Convolution models."""

import re

from numpy import convolve, exp, log, linspace

from ufit.param import Param
from ufit.models.base import Model

__all__ = ['GaussianConvolution']


id_re = re.compile('[a-zA-Z][a-zA-Z0-9_]*$')

class GaussianConvolution(Model):
    """Models a 1-D convolution with a Gaussian kernel.

    Parameters:

    * `width` - FWHM of Gaussian kernel
    """

    def __init__(self, model, width=1, name=None):
        self._model = model
        if name is not None:
            self.name = name
        elif model.name and id_re.match(model.name):
            self.name = '%s_conv' % model.name
        else:
            self.name = 'conv'
        self.params = model.params[:]
        pname = self.name + '_width'
        self.params.append(Param.from_init(pname, width))

        def convfcn(p, x):
            data = model.fcn(p, x)
            # construct Gaussian filter
            N = len(x)   # number of data points
            M = 2*N - 1  # number of filter points
            # "valid" convolution mode now returns M - N + 1 == N points
            binwidth = (x.max() - x.min()) / len(x)
            filtx = linspace(-N*binwidth, N*binwidth, M)
            filt = exp(-filtx**2 / p[pname]**2 * 4*log(2))
            # normalize filter
            filt /= filt.sum()
            return convolve(data, filt, mode='valid')
        self.fcn = convfcn



#  -*- coding: utf-8 -*-
# *****************************************************************************
# ufit, a universal scattering fitting suite
#
# Copyright (c) 2013-2014, Georg Brandl and contributors.  All rights reserved.
# Licensed under a 2-clause BSD license, see LICENSE.
# *****************************************************************************

"""Model of S(q,w) with resolution convolution for TAS."""

import time
import inspect

from numpy import matrix as zeros

from ufit.rescalc import resmat, calc_MC, calc_MC_cluster, calc_MC_mcstas, \
    load_cfg, load_par, PARNAMES, CFGNAMES, plot_resatpoint
from ufit.models.base import Model
from ufit.param import prepare_params, update_params

__all__ = ['ConvolvedScatteringLaw']


class ConvolvedScatteringLaw(Model):
    """Model using a scattering law given as a function and a set of resolution
    parameters for a triple-axis spectrometer.

    Signature of the S(q,w) function given by *sqw*::

       sqw(h, k, l, E, QE0, Sigma, par0, par1, ...)

    If the function is given as a string, it it taken as ``module:function``.
    The calculation is then clustered using SSH and ufit.cluster.

    h,k,l,E are arrays of the Monte-Carlo point coordinates, QE0 is the center
    of the ellipse, Sigma are the ellipse widths around the center.  A constant
    background is always included as a parameter named "bkgd".

    *instfiles* must be ('instr.cfg', 'instr.par').

    Initial parameters can include parameters named cfg_... and par_...; this
    will add instrument configuration and parameters to the fitting parameters.
    They can be given initial values or a None value, in which case the initial
    value will come from the .cfg or .par file.

    Use cfg_ALL=1 or par_ALL=1 to add all cfg or par entries to the fit
    parameters (this is mostly useful to interactively play around with the
    resolution in one scan).
    """
    nsamples = -1  # for plotting: plot only 4x as many points as datapoints

    def __init__(self, sqw, instfiles, NMC=2000, name=None, cluster=False,
                 mcstas=None, matrix=None, mathkl=None, **init):
        self._cluster = False
        self._mcstas = mcstas
        if isinstance(sqw, str):
            modname, funcname = sqw.split(':')
            mod = __import__(modname)
            code = open(mod.__file__.rstrip('c')).read()
            self._sqwfunc = funcname
            self._sqwcode = code
            self._sqw = getattr(mod, funcname)
            self.name = funcname
            self._cluster = cluster
        else:  # cannot cluster
            self._sqw = sqw
            self.name = name or sqw.__name__
        init['NMC'] = str(NMC)  # str() makes it a fixed parameter

        instparnames = []
        par_all = False
        cfg_all = False
        self._instpars = []

        cfg_orig = load_cfg(instfiles[0])
        par_orig = load_par(instfiles[1])

        for par in init:
            if par == 'par_ALL':
                par_all = True
            elif par == 'cfg_ALL':
                cfg_all = True
            elif par.startswith('par_'):
                if par[4:] in PARNAMES:
                    self._instpars.append(par[4:])
                    instparnames.append(par)
                    if init[par] is None:
                        init[par] = str(par_orig[par[4:]])
                else:
                    raise Exception('invalid instrument parameter: %r' % par)
            elif par.startswith('cfg_'):
                if par[4:] in CFGNAMES:
                    self._instpars.append(CFGNAMES.index(par[4:]))
                    instparnames.append(par)
                    if init[par] is None:
                        init[par] = str(cfg_orig[CFGNAMES.index(par[4:])])
                else:
                    raise Exception('invalid instrument configuration: %r' % par)

        if par_all:
            instparnames = [pn for pn in instparnames if not pn.startswith('par_')]
            self._instpars = [pn for pn in self._instpars if not isinstance(pn, str)]
            instparnames.extend('par_' + pn for pn in PARNAMES)
            self._instpars.extend(PARNAMES)
            for pn in PARNAMES:
                init['par_' + pn] = str(par_orig[pn])
        if cfg_all:
            instparnames = [pn for pn in instparnames if not pn.startswith('cfg_')]
            self._instpars = [ip for ip in self._instpars if isinstance(ip, str)]
            instparnames.extend('cfg_' + x for x in CFGNAMES)
            self._instpars.extend(range(len(CFGNAMES)))
            for i in range(len(CFGNAMES)):
                init['cfg_' + CFGNAMES[i]] = str(cfg_orig[i])

        # numba compat
        arg_sqw = getattr(self._sqw, 'py_func', self._sqw)
        self._pvs = self._init_params(name,
            ['NMC', 'bkgd'] + inspect.getargspec(arg_sqw)[0][6:] +
            instparnames, init)
        self._ninstpar = len(instparnames)
        self._resmat = resmat(cfg_orig, par_orig)

        if matrix is not None:
            self._resmat.fixed_res = True
            self._resmat.setNPMatrix(matrix, mathkl)
            #self._resmat.NP = matrix
            self._resmat.R0_corrected = 1.0

    def fcn(self, p, x):
        parvalues = [p[pv] for pv in self._pvs]
        t1 = time.time()
        #print 'Sqw: values = ', parvalues
        if self._ninstpar:
            sqwpar  = parvalues[2:-self._ninstpar]
            for pn, pv in zip(self._instpars, parvalues[-self._ninstpar:]):
                if isinstance(pn, str):
                    self._resmat.par[pn] = pv
                else:
                    self._resmat.cfg[pn] = pv
            use_caching = False
        else:
            sqwpar = parvalues[2:]
            use_caching = True
        if self._mcstas:
            res = calc_MC_mcstas(x, sqwpar, self._sqw, self._resmat,
                                 parvalues[0])
        elif self._cluster:
            res = calc_MC_cluster(x, sqwpar, self._sqwcode,
                                  self._sqwfunc, self._resmat, parvalues[0],
                                  use_caching=use_caching)
        else:
            res = calc_MC(x, sqwpar, self._sqw, self._resmat,
                          parvalues[0], use_caching=use_caching)
        res += parvalues[1]  # background
        t2 = time.time()
        #print 'Sqw: iteration = %.3f sec' % (t2-t1)
        return res

    def resplot(self, h, k, l, e):
        self._resmat.sethklen(h, k, l, e)
        plot_resatpoint(self._resmat.cfg, self._resmat.par, self._resmat)

    def simulate(self, data):
        varying, varynames, dependent, _ = prepare_params(self.params, data.meta)
        pd = dict((p.name, p.value) for p in self.params)
        update_params(dependent, data.meta, pd)
        yy = self.fcn(pd, data.x)
        new = data.copy()
        new.y = yy
        new.dy = zeros(len(yy))
        return new



#  -*- coding: utf-8 -*-
# *****************************************************************************
# ufit, a universal scattering fitting suite
#
# Copyright (c) 2013-2014, Georg Brandl and contributors.  All rights reserved.
# Licensed under a 2-clause BSD license, see LICENSE.
# *****************************************************************************

"""Backend using pyminuit."""

from __future__ import absolute_import

from ufit.param import prepare_params, update_params
from ufit.utils import get_chisqr

try:
    from minuit import Minuit
except ImportError:
    ## minuit2 crashes with GIL problems here
    #from minuit2 import Minuit2 as Minuit
    raise

__all__ = ['do_fit', 'backend_name']

backend_name = 'minuit'

def do_fit(data, fcn, params, add_kw):
    x, y, dy = data.fit_columns
    meta = data.meta
    varying, varynames, dependent, _ = prepare_params(params, meta)

    # sadly, parameter names are restricted to 10 characters with pyminuit
    minuitnames = ['p%d' % j for j in range(len(varynames))]
    minuit_map = dict(zip(varynames, minuitnames))

    # also sadly, pyminuit insists on a function with the exact number and
    # names of the parameters in the signature, so we have to create
    # such a function dynamically

    code = 'def minuitfcn(' + ', '.join(minuitnames) + '''):
        pd = {''' + ', '.join("%r: %s" % v for v in minuit_map.items()) + '''}
        update_params(dependent, meta, pd)
        return ((fcn(pd, x) - y)**2 / dy**2).sum()
    '''

    fcn_environment = {'meta': meta, 'x': x, 'y': y, 'dy': dy,
                       'fcn': fcn, 'dependent': dependent}
    fcn_environment['update_params'] = update_params   # it's a global
    exec code in fcn_environment

    m = Minuit(fcn_environment['minuitfcn'])
    m.up = 1.0
    for kw in add_kw:
        setattr(m, kw, add_kw[kw])
    for p in varying:
        m.values[minuit_map[p.name]] = p.value
        m.errors[minuit_map[p.name]] = 100*p.delta or p.value/100. or 0.01
        if p.pmin is not None or p.pmax is not None:
            m.limits[minuit_map[p.name]] = (p.pmin is None and -1e8 or p.pmin,
                                            p.pmax is None and +1e8 or p.pmax)
    try:
        m.migrad()
        m.hesse()
    except Exception, e:
        return False, str(e), 0
    #m.minos()  -> would calculate more exact and asymmetric errors

    pd = dict((pn, m.values[minuit_map[pn]]) for pn in varynames)
    update_params(dependent, meta, pd)
    for p in params:
        p.value = pd[p.name]
        if p.name in minuit_map:
            p.error = m.errors[minuit_map[p.name]]
            p.correl = {}  # XXX
        else:
            p.error = 0
            p.correl = {}

    return True, '', get_chisqr(fcn, x, y, dy, params)



#  -*- coding: utf-8 -*-
# *****************************************************************************
# ufit, a universal scattering fitting suite
#
# Copyright (c) 2013-2014, Georg Brandl and contributors.  All rights reserved.
# Licensed under a 2-clause BSD license, see LICENSE.
# *****************************************************************************

"""Session item for datasets and corresponding GUI."""

from os import path

from numpy import savetxt, array, linspace, sqrt, mean

from PyQt4.QtCore import pyqtSignature as qtsig, SIGNAL
from PyQt4.QtGui import QTabWidget, QWidget, QDialog, QMessageBox

from ufit.data.merge import rebin
from ufit.param import prepare_params
from ufit.models import eval_model
from ufit.gui import logger
from ufit.gui.modelbuilder import ModelBuilder
from ufit.gui.dataops import DataOps
from ufit.gui.fitter import Fitter
from ufit.gui.session import session, SessionItem
from ufit.gui.mappingitem import MappingItem
from ufit.gui.common import loadUi
from ufit.gui.dialogs import ParamSetDialog


def default_model(data):
    ymin = data.y.min()
    ymaxidx = data.y.argmax()
    ymax = data.y[ymaxidx]
    xmax = data.x[ymaxidx]
    overhalf = data.x[data.y > (ymax + ymin)/2.]
    if len(overhalf) >= 2:
        xwidth = abs(overhalf[0] - overhalf[-1]) or 0.1
    else:
        xwidth = 0.1
    new_model = eval_model('Background() + Gauss(\'peak\')')
    new_model.params[0].value = ymin
    new_model.params[1].value = xmax
    new_model.params[2].value = ymax-ymin
    new_model.params[3].value = xwidth
    return new_model


class ScanDataItem(SessionItem):

    itemtype = 'scan'

    def __init__(self, data, model=None):
        self.data = data
        self.model = model or default_model(data)
        SessionItem.__init__(self)

    def change_model(self, model):
        self.model = model
        self.emit(SIGNAL('newModel'), model)
        session.set_dirty()

    def after_load(self):
        self.data.after_load()  # upgrade datastructures

    def __reduce__(self):
        return (self.__class__, (self.data, self.model))

    def create_panel(self, mainwindow, canvas):
        return ScanDataPanel(mainwindow, canvas, self)

    def create_multi_panel(self, mainwindow, canvas):
        return MultiDataOps(mainwindow, canvas)

    def update_htmldesc(self):
        title = self.data.title
        # XXX self.dataops.titleEdit.setText(title)
        self.title = title
        self.htmldesc = '<big><b>%s</b></big>' % self.index + \
            (title and ' - %s' % title or '') + \
            (self.data.environment and
             '<br>%s' % ', '.join(self.data.environment) or '') + \
            ('<br><small>%s</small>' % '<br>'.join(self.data.sources))
        session.emit(SIGNAL('itemsUpdated'))

    def export_python(self, filename):
        with open(filename, 'wb') as fp:
            fp.write('from ufit.lab import *\n')
            fp.write('\n')
            self.data.export_python(fp, 'data')
            fp.write('\n')
            self.model.export_python(fp, 'model')
            fp.write('''\
## just plot current values
data.plot()
model.plot_components(data)
model.plot(data)

## to fit again use this...
#result = model.fit(data)
#result.printout()
#result.plot()

show()
''')

    def export_ascii(self, filename):
        with open(filename, 'wb') as fp:
            self.data.export_ascii(fp)

    def export_fits(self, filename):
        xx = linspace(self.data.x.min(), self.data.x.max(), 1000)
        paramvalues = prepare_params(self.model.params, self.data.meta)[3]
        yy = self.model.fcn(paramvalues, xx)
        yys = []
        for comp in self.model.get_components():
              if comp is self.model:
                  continue
              yys.append(comp.fcn(paramvalues, xx))
        savetxt(filename, array([xx, yy] + yys).T)


class ScanDataPanel(QTabWidget):
    def __init__(self, parent, canvas, item):
        QTabWidget.__init__(self, parent)
        self.item = item
        self.dataops = DataOps(self)
        self.mbuilder = ModelBuilder(self)
        self.fitter = Fitter(self)
        self._limits = None
        self._dont_update_modeldef = False
        self.picker_widget = None

        self.connect(self.item, SIGNAL('newModel'), self.on_item_newModel)

        self.canvas = canvas
        self.dataops.initialize(item)
        self.mbuilder.initialize(item.data, item.model)
        self.fitter.initialize(item.model, item.data, fit=False)
        self.connect(self.dataops, SIGNAL('pickRequest'), self.set_picker)
        self.connect(self.dataops, SIGNAL('replotRequest'), self.plot)
        self.connect(self.dataops, SIGNAL('titleChanged'), self.item.update_htmldesc)
        self.connect(self.mbuilder, SIGNAL('newModel'), self.on_mbuilder_newModel)
        self.connect(self.mbuilder, SIGNAL('pickRequest'), self.set_picker)
        self.connect(self.fitter, SIGNAL('replotRequest'), self.plot)
        self.connect(self.fitter, SIGNAL('pickRequest'), self.set_picker)
        self.addTab(self.dataops, 'Data operations')
        self.addTab(self.mbuilder, 'Modeling')
        self.addTab(self.fitter, 'Fitting')
        self.setCurrentWidget(self.mbuilder)

    def on_mbuilder_newModel(self, model, update_modeldef=False,
                             switch_fitter=True):
        self._dont_update_modeldef = not update_modeldef
        try:
            self.item.change_model(model)
        finally:
            self._dont_update_modeldef = False
        if switch_fitter:
            self.setCurrentWidget(self.fitter)

    def on_item_newModel(self, model):
        if not self._dont_update_modeldef:
            self.mbuilder.modeldefEdit.setText(model.get_description())
        self.fitter.initialize(model, self.item.data, fit=False, keep_old=True)

    def set_picker(self, widget):
        self.picker_widget = widget

    def on_canvas_pick(self, event):
        if self.picker_widget:
            self.picker_widget.on_canvas_pick(event)

    def save_limits(self):
        self._limits = self.canvas.axes.get_xlim(), self.canvas.axes.get_ylim()

    def get_saved_limits(self):
        return self._limits

    def plot(self, limits=True, canvas=None):
        canvas = canvas or self.canvas
        plotter = canvas.plotter
        plotter.reset(limits)
        try:
            plotter.plot_data(self.item.data)
            plotter.plot_model_full(self.item.model, self.item.data)
        except Exception:
            logger.exception('Error while plotting')
        else:
            canvas.draw()

    def export_ascii(self, filename):
        self.item.export_ascii(filename)

    def export_fits(self, filename):
        self.item.export_fits(filename)

    def export_python(self, filename):
        self.item.export_python(filename)


class MultiDataOps(QWidget):

    def __init__(self, parent, canvas):
        QWidget.__init__(self, parent)
        self.canvas = canvas
        self.connect(self, SIGNAL('replotRequest'), self.plot)

        loadUi(self, 'multiops.ui')

    def initialize(self, items):
        self.items = [i for i in items if isinstance(i, ScanDataItem)]
        self.datas = [i.data for i in self.items]
        self.monscaleEdit.setText(str(int(mean([d.nscale for d in self.datas]))))
        self.onemodelBox.clear()
        self.onemodelBox.addItems(['%d' % i.index for i in self.items])

    def plot(self, limits=True, canvas=None):
        canvas = canvas or self.canvas
        xlabels = set()
        ylabels = set()
        titles = set()
        canvas.plotter.reset()
        for i in self.items:
            c = canvas.plotter.plot_data(i.data, multi=True)
            canvas.plotter.plot_model(i.model, i.data, labels=False, color=c)
            xlabels.add(i.data.xaxis)
            ylabels.add(i.data.yaxis)
            titles.add(i.data.title)
        canvas.plotter.plot_finish(', '.join(xlabels), ', '.join(ylabels),
                                   ', '.join(titles).decode('ascii', 'ignore'))
        canvas.draw()

    @qtsig('')
    def on_rebinBtn_clicked(self):
        try:
            binsize = float(self.precisionEdit.text())
        except ValueError:
            QMessageBox.warning(self, 'Error', 'Enter a valid precision.')
            return
        for data in self.datas:
            new_array = rebin(data._data, binsize)
            data.__init__(data.meta, new_array,
                          data.xcol, data.ycol, data.ncol,
                          data.nscale, name=data.name,
                          sources=data.sources)
        self.emit(SIGNAL('replotRequest'), None)
        session.set_dirty()

    @qtsig('')
    def on_mulBtn_clicked(self):
        try:
            const = float(self.scaleConstEdit.text())
        except ValueError:
            return
        for data in self.datas:
            data.y *= const
            data.y_raw *= const
            data.dy *= const
            data.dy_raw *= const
        self.emit(SIGNAL('replotRequest'), None)
        session.set_dirty()

    @qtsig('')
    def on_addBtn_clicked(self):
        try:
            const = float(self.addConstEdit.text())
        except ValueError:
            return
        for data in self.datas:
            data.y += const
            data.y_raw += const * data.norm
        self.emit(SIGNAL('replotRequest'), None)
        session.set_dirty()

    @qtsig('')
    def on_scaleXBtn_clicked(self):
        try:
            const = float(self.scaleXConstEdit.text())
        except ValueError:
            return
        for data in self.datas:
            data.x *= const
        self.emit(SIGNAL('replotRequest'), None)
        session.set_dirty()

    @qtsig('')
    def on_shiftBtn_clicked(self):
        try:
            const = float(self.shiftConstEdit.text())
        except ValueError:
            return
        for data in self.datas:
            data.x += const
        self.emit(SIGNAL('replotRequest'), None)
        session.set_dirty()

    @qtsig('')
    def on_monscaleBtn_clicked(self):
        try:
            const = int(self.monscaleEdit.text())
        except ValueError:
            return
        for data in self.datas:
            data.nscale = const
            data.norm = data.norm_raw / const
            data.y = data.y_raw/data.norm
            data.dy = sqrt(data.y_raw)/data.norm
            data.yaxis = data.ycol + ' / %s %s' % (const, data.ncol)
        self.emit(SIGNAL('replotRequest'), None)
        session.set_dirty()

    @qtsig('')
    def on_mergeBtn_clicked(self):
        try:
            precision = float(self.mergeEdit.text())
        except ValueError:
            QMessageBox.warning(self, 'Error', 'Enter a valid precision.')
            return
        new_data = self.datas[0].merge(precision, *self.datas[1:])
        session.add_item(ScanDataItem(new_data), self.items[-1].group)

    @qtsig('')
    def on_floatMergeBtn_clicked(self):
        try:
            precision = float(self.mergeEdit.text())
        except ValueError:
            QMessageBox.warning(self, 'Error', 'Enter a valid precision.')
            return
        new_data = self.datas[0].merge(precision, floatmerge=True, *self.datas[1:])
        session.add_item(ScanDataItem(new_data), self.items[-1].group)

    @qtsig('')
    def on_onemodelBtn_clicked(self):
        which = self.onemodelBox.currentIndex()
        if which < 0:
            return
        model = self.items[which].model
        for i, item in enumerate(self.items):
            if i == which:
                continue
            item.change_model(model.copy())
        self.emit(SIGNAL('replotRequest'), None)

    @qtsig('')
    def on_fitallBtn_clicked(self):
        for item in self.items:
            res = item.model.fit(item.data)
            session.emit(SIGNAL('modelFitted'), item, res)
        self.emit(SIGNAL('replotRequest'), None)

    @qtsig('')
    def on_paramsetBtn_clicked(self):
        dlg = ParamSetDialog(self, self.items)
        if dlg.exec_() != QDialog.Accepted:
            return
        session.add_item(ScanDataItem(dlg.new_data), self.items[-1].group)

    @qtsig('')
    def on_mappingBtn_clicked(self):
        item = MappingItem([item.data for item in self.items], None)
        session.add_item(item, self.items[-1].group)

    @qtsig('')
    def on_globalfitBtn_clicked(self):
        QMessageBox.warning(self, 'Sorry', 'Not implemented yet.')

    def export_ascii(self, filename):
        base, ext = path.splitext(filename)
        for i, item in enumerate(self.items):
            item.export_ascii(base + '.%d' % i + ext)

    def export_fits(self, filename):
        base, ext = path.splitext(filename)
        for i, item in enumerate(self.items):
            item.export_fits(base + '.%d' % i + ext)

    def export_python(self, filename):
        base, ext = path.splitext(filename)
        for i, item in enumerate(self.items):
            item.export_python(base + '.%d' % i + ext)



#  -*- coding: utf-8 -*-
# *****************************************************************************
# ufit, a universal scattering fitting suite
#
# Copyright (c) 2013-2014, Georg Brandl and contributors.  All rights reserved.
# Licensed under a 2-clause BSD license, see LICENSE.
# *****************************************************************************

"""Utility functions and classes."""

import re
from os import path


def get_chisqr(fcn, x, y, dy, params):
    paramvalues = dict((p.name, p.value) for p in params)
    sum_sqr = ((fcn(paramvalues, x) - y)**2 / dy**2).sum()
    nfree = len(y) - sum(1 for p in params if not p.expr)
    return sum_sqr / nfree


class attrdict(dict):
    def __getattr__(self, key):
        try:
            return self[key]
        except KeyError:
            raise AttributeError(key)
    def __setattr__(self, key, value):
        self[key] = value
    def __delattr__(self, key):
        try:
            del self[key]
        except KeyError:
            raise AttributeError(key)


_missing = object()

class cached_property(object):
    """A decorator that converts a function into a lazy property.  The
    function wrapped is called the first time to retrieve the result
    and then that calculated result is used the next time you access
    the value.
    """

    def __init__(self, func, name=None, doc=None):
        self.__name__ = name or func.__name__
        self.__module__ = func.__module__
        self.__doc__ = doc or func.__doc__
        self.func = func

    def __get__(self, obj, type=None):
        if obj is None:
            return self
        value = obj.__dict__.get(self.__name__, _missing)
        if value is _missing:
            value = self.func(obj)
            obj.__dict__[self.__name__] = value
        return value


numor_re = re.compile(r'\d+')

def extract_template(fn):
    bn = path.basename(fn)
    dn = path.dirname(fn)
    m = list(numor_re.finditer(bn))
    if not m:
        dtempl = fn
        numor = 0
    else:
        b, e = m[-1].span()
        dtempl = path.join(dn, bn[:b] + '%%0%dd' % (e-b) + bn[e:])
        numor = int(m[-1].group())
    return dtempl, numor



#  -*- coding: utf-8 -*-
# *****************************************************************************
# ufit, a universal scattering fitting suite
#
# Copyright (c) 2013-2014, Georg Brandl and contributors.  All rights reserved.
# Licensed under a 2-clause BSD license, see LICENSE.
# *****************************************************************************

"""Data browsing window for the standalone GUI."""

import os
from os import path

from PyQt4.QtCore import pyqtSignature as qtsig, QByteArray
from PyQt4.QtGui import QMainWindow, QApplication, QListWidgetItem, \
    QVBoxLayout, QFileDialog

from ufit.data import Loader
from ufit.utils import extract_template
from ufit.gui import logger
from ufit.gui.common import loadUi, MPLCanvas, MPLToolbar, SettingGroup, \
    path_to_str
from ufit.gui.session import session
from ufit.gui.scanitem import ScanDataItem


class BrowseWindow(QMainWindow):
    def __init__(self, parent):
        QMainWindow.__init__(self, parent)
        loadUi(self, 'browse.ui')
        self.logger = logger.getChild('browse')

        self.rootdir = ''
        self.loader = Loader()
        self._data = {}
        self.canvas = MPLCanvas(self)
        self.canvas.plotter.lines = True
        self.toolbar = MPLToolbar(self.canvas, self)
        self.toolbar.setObjectName('browsetoolbar')
        self.addToolBar(self.toolbar)
        layout = QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self.canvas)
        self.plotframe.setLayout(layout)
        self.sgroup = SettingGroup('browse')

        with self.sgroup as settings:
            geometry = settings.value('geometry', QByteArray())
            self.restoreGeometry(geometry)
            windowstate = settings.value('windowstate', QByteArray())
            self.restoreState(windowstate)
            splitstate = settings.value('splitstate', QByteArray())
            self.splitter.restoreState(splitstate)
            #vsplitstate = settings.value('vsplitstate').toByteArray()
            #self.vsplitter.restoreState(vsplitstate)

    @qtsig('')
    def on_loadBtn_clicked(self):
        datas = [self._data[item.type()] for item in self.dataList.selectedItems()]
        if not datas:
            return
        items = [ScanDataItem(data) for data in datas]
        # XXX which group
        session.add_items(items)

    @qtsig('')
    def on_dirBtn_clicked(self):
        newdir = QFileDialog.getExistingDirectory(self, 'New directory',
                                                  self.rootdir)
        self.set_directory(path_to_str(newdir))

    @qtsig('')
    def on_refreshBtn_clicked(self):
        self.set_directory(self.rootdir)

    def set_directory(self, root):
        self.setWindowTitle('ufit browser - %s' % root)
        self.canvas.axes.text(0.5, 0.5, 'Please wait, loading all data...',
                              horizontalalignment='center')
        self.canvas.draw()
        QApplication.processEvents()
        self.rootdir = root
        files = os.listdir(root)
        self.dataList.clear()
        for fn in sorted(files):
            fn = path.join(root, fn)
            if not path.isfile(fn):
                continue
            try:
                t, n = extract_template(fn)
                self.loader.template = t
                res = self.loader.load(n, 'auto', 'auto', 'auto', 'auto', -1)
            except Exception, e:
                self.logger.warning('While loading %r: %s' % (fn, e))
            else:
                self._data[n] = res
                QListWidgetItem('%s (%s) - %s - %s' %
                                (n, res.xcol, res.title,
                                 ', '.join(res.environment)),
                                self.dataList, n)
        self.canvas.axes.clear()
        self.canvas.draw()

    def on_dataList_itemSelectionChanged(self):
        numors = [item.type() for item in self.dataList.selectedItems()]
        if not numors:
            return
        plotter = self.canvas.plotter
        plotter.reset(False)
        if len(numors) > 1:
            for numor in numors:
                plotter.plot_data(self._data[numor], multi=True)
            plotter.plot_finish()
        else:
            plotter.plot_data(self._data[numors[0]])
        plotter.draw()

    def closeEvent(self, event):
        event.accept()
        with self.sgroup as settings:
            settings.setValue('geometry', self.saveGeometry())
            settings.setValue('windowstate', self.saveState())
            settings.setValue('splitstate', self.splitter.saveState())
            #settings.setValue('vsplitstate',
            #                  QVariant(self.vsplitter.saveState()))



#  -*- coding: utf-8 -*-
# *****************************************************************************
# ufit, a universal scattering fitting suite
#
# Copyright (c) 2013-2014, Georg Brandl and contributors.  All rights reserved.
# Licensed under a 2-clause BSD license, see LICENSE.
# *****************************************************************************

"""Logging utilities for ufit.

Mostly borrowed from nicos.utils.loggers.
"""

import os
import time
import traceback
from os import path
from logging import Formatter, Handler, StreamHandler, DEBUG, INFO, WARNING, \
    root, getLogger

__all__ = ['getLogger']

DATEFMT = '%H:%M:%S'

# console color utils

_codes = {}

_attrs = {
    'reset':     '39;49;00m',
    'bold':      '01m',
    'faint':     '02m',
    'standout':  '03m',
    'underline': '04m',
    'blink':     '05m',
}

for _name, _value in _attrs.items():
    _codes[_name] = '\x1b[' + _value

_colors = [
    ('black', 'darkgray'),
    ('darkred', 'red'),
    ('darkgreen', 'green'),
    ('brown', 'yellow'),
    ('darkblue', 'blue'),
    ('purple', 'fuchsia'),
    ('turquoise', 'teal'),
    ('lightgray', 'white'),
]

for _i, (_dark, _light) in enumerate(_colors):
    _codes[_dark] = '\x1b[%im' % (_i + 30)
    _codes[_light] = '\x1b[%i;01m' % (_i + 30)

def colorize(name, text):
    return _codes.get(name, '') + text + _codes.get('reset', '')

def colorcode(name):
    return _codes.get(name, '')

def nocolor():
    for key in list(_codes):
        _codes[key] = ''

if os.name == 'nt':
    try:
        # colorama provides ANSI-colored console output support under Windows
        import colorama  # pylint: disable=F0401
    except ImportError:
        nocolor()
    else:
        colorama.init()


class ConsoleFormatter(Formatter):
    """
    A lightweight formatter for the interactive console, with optional
    colored output.
    """

    def __init__(self, fmt=None, datefmt=None, colorize=None):
        Formatter.__init__(self, fmt, datefmt)
        if colorize:
            self.colorize = colorize
        else:
            self.colorize = lambda c, s: s

    def formatException(self, exc_info):
        return ''.join(traceback.format_exception(*exc_info))

    def formatTime(self, record, datefmt=None):
        return time.strftime(datefmt or DATEFMT,
                             self.converter(record.created))

    def format(self, record):
        levelno = record.levelno
        datefmt = self.colorize('lightgray', '[%(asctime)s] ')
        record.message = record.getMessage()
        if record.name == 'ufit':
            namefmt = ''
        else:
            namefmt = '%(name)s: '
        if levelno <= DEBUG:
            fmtstr = self.colorize('darkgray', '%s%%(message)s' % namefmt)
        elif levelno <= INFO:
            fmtstr = '%s%%(message)s' % namefmt
        elif levelno <= WARNING:
            fmtstr = self.colorize('fuchsia', '%s%%(levelname)s: %%(message)s'
                                   % namefmt)
        else:
            fmtstr = self.colorize('red', '%s%%(levelname)s: %%(message)s'
                                   % namefmt)
        fmtstr = datefmt + fmtstr
        record.asctime = self.formatTime(record, self.datefmt)
        s = fmtstr % record.__dict__
        if record.exc_info:
            # *not* caching exception text on the record, since it's
            # only a short version
            s += '\n' + self.formatException(record.exc_info).rstrip()
        return s


LOGFMT = '%(asctime)s : %(levelname)-7s : %(name)s: %(message)s'
DATESTAMP_FMT = '%Y-%m-%d'
SECONDS_PER_DAY = 60 * 60 * 24

class LogfileHandler(StreamHandler):
    """
    Logs to log files with a date stamp appended, and rollover on midnight.
    """

    def __init__(self, directory, filenameprefix='ufit', dayfmt=DATESTAMP_FMT):
        if not path.isdir(directory):
            os.makedirs(directory)
        self._currentsymlink = path.join(directory, 'current')
        self._pathnameprefix = path.join(directory, filenameprefix)
        self._dayfmt = dayfmt
        # today's logfile name
        basefn = self._pathnameprefix + '-' + time.strftime(dayfmt) + '.log'
        self.baseFilename = path.abspath(basefn)
        self.stream = None
        Handler.__init__(self)   # don't open file yet
        # determine time of first midnight from now on
        t = time.localtime()
        self.rollover_at = time.mktime((t[0], t[1], t[2], 0, 0, 0,
                                        t[6], t[7], t[8])) + SECONDS_PER_DAY
        self.setFormatter(Formatter(LOGFMT, DATEFMT))

    def _open(self):
        # update 'current' symlink upon open
        try:
            os.remove(self._currentsymlink)
        except OSError:
            # if the symlink does not (yet) exist, OSError is raised.
            # should happen at most once per installation....
            pass
        if hasattr(os, 'symlink'):
            os.symlink(path.basename(self.baseFilename), self._currentsymlink)
        # finally open the new logfile....
        return open(self.baseFilename, 'a')

    def emit(self, record):
        try:
            t = int(time.time())
            if t >= self.rollover_at:
                self.doRollover()
            if self.stream is None:
                self.stream = self._open()
            StreamHandler.emit(self, record)
        except Exception:
            self.handleError(record)

    def close(self):
        self.acquire()
        try:
            if self.stream:
                self.flush()
                self.stream.close()
                self.stream = None
                Handler.close(self)
        finally:
            self.release()

    def doRollover(self):
        if self.stream:
            self.stream.close()
            self.stream = None
        self.baseFilename = self._pathnameprefix + '-' + \
                            time.strftime(self._dayfmt) + '.log'
        self.rollover_at += SECONDS_PER_DAY


handler = StreamHandler()
fmter = ConsoleFormatter(colorize=colorize)
handler.setFormatter(fmter)
root.addHandler(handler)
loghandler = LogfileHandler(path.expanduser('~/.config/ufit'))
loghandler.setLevel(WARNING)
root.addHandler(loghandler)
root.setLevel(INFO)
